from pathlib import Path
from typing import Iterator

from whispers.models.pair import KeyValuePair


class Gradle:
    def pairs(self, filepath: Path) -> Iterator[KeyValuePair]:
        key = "password"

        for lineno, line in enumerate(filepath.open(), 1):
            line = line.strip()

            if key not in line:
                continue

            value = line.split(key)[-1].strip(":) ")
            if value:
                yield KeyValuePair(key, value, line=lineno)



from pathlib import Path
from typing import Iterator

from whispers.models.pair import KeyValuePair


class Pypirc:
    def pairs(self, filepath: Path) -> Iterator[KeyValuePair]:
        for lineno, line in enumerate(filepath.open(), 1):
            if "password:" not in line:
                continue

            value = line.split("password:")[-1].strip()
            if value:
                key = "pypi password"
                yield KeyValuePair(key, value, line=lineno)



from pathlib import Path
from typing import Iterator

from jproperties import Properties

from whispers.models.pair import KeyValuePair


class Jproperties:
    def pairs(self, filepath: Path) -> Iterator[KeyValuePair]:
        props = Properties()
        props.load(filepath.read_text(), "utf-8")
        for key, value in props.properties.items():
            key = key.replace(".", "_")
            yield KeyValuePair(key, value, [key])



import logging
from argparse import Namespace
from typing import List

from whispers.core.utils import default_rules
from whispers.models.appconfig import AppConfig
from whispers.models.rule import Rule


def load_rules(args: Namespace, config: AppConfig) -> List[dict]:
    """Loads applicable rules based on args and config"""
    include_rule_ids = args.rules or config.include.rules
    exclude_rule_ids = args.xrules or config.exclude.rules
    include_groups = args.groups or config.include.groups
    exclude_groups = args.xgroups or config.exclude.groups
    include_severity = args.severity or config.include.severity
    exclude_severity = args.xseverity or config.exclude.severity
    applicable_rules = []

    # Load from default rules based on rules/severity config
    for rule in default_rules():
        rule_id = rule.get("id", None)
        rule_group = rule.get("group", None)
        rule_severity = rule.get("severity", None)

        if rule_id in exclude_rule_ids:
            continue  # Rule excluded

        if rule_id not in include_rule_ids:
            continue  # Rule not included

        if rule_group in exclude_groups:
            continue  # Group excluded

        if rule_group not in include_groups:
            continue  # Group not included

        if rule_severity in exclude_severity:
            continue  # Severity excluded

        if rule_severity not in include_severity:
            continue  # Severity not included

        applicable_rules.append(Rule(rule))

    # Load inline rules from config file (if any)
    for rule in include_rule_ids:
        if isinstance(rule, str):
            continue  # Not an inline rule

        applicable_rules.append(Rule(rule))

    logging.debug(f"load_rules loaded {len(applicable_rules)} rules")
    return applicable_rules



from pathlib import Path
from typing import Iterator, Tuple

import astroid

from whispers.core.utils import global_exception_handler
from whispers.models.pair import KeyValuePair


class Python:
    """
    Parses Python source code into AST using Astroid
    http://pylint.pycqa.org/projects/astroid/en/latest/api/astroid.nodes.html#id1
    Returns key-value pairs
    """

    def __init__(self):
        self.keypath = []

    def pairs(self, filepath: Path) -> Iterator[KeyValuePair]:
        try:
            code = filepath.read_text()
            tree = astroid.parse(code)
            yield from self.traverse(tree)
        except Exception:
            global_exception_handler(filepath.as_posix(), code[:100])

    @staticmethod
    def is_key(node) -> bool:
        """
        Check if node is a variable
        """
        types = [astroid.nodes.Name, astroid.nodes.AssignName]
        return type(node) in types

    @staticmethod
    def is_value(node) -> bool:
        """
        Check if node is a value
        """
        types = [astroid.nodes.Const, astroid.nodes.JoinedStr, astroid.nodes.Call]
        return type(node) in types

    @staticmethod
    def node_concat_const(nodes) -> str:
        """
        Joins all Const node values and returns as string
        Return empty string if a non-Const node is found
        """
        ret = ""
        for value in nodes:
            if not isinstance(value, astroid.nodes.Const):
                return ""
            ret += f"{value.value}"
        return ret

    def node_to_str(self, node) -> str:
        """
        Converts node valid objects to string
        Returns empty string otherwise
        """
        if isinstance(node, astroid.nodes.Name):
            return node.name
        if isinstance(node, astroid.nodes.AssignName):
            return node.name
        if isinstance(node, astroid.nodes.Const):
            return node.value
        if isinstance(node, astroid.nodes.JoinedStr):
            return self.node_concat_const(node.values)
        if isinstance(node, astroid.nodes.Keyword):
            return node.arg
        if isinstance(node, astroid.nodes.Call):
            return self.node_concat_const(node.args)
        return ""

    def traverse(self, tree) -> Iterator[Tuple]:
        """
        Recursively traverse nodes yielding key-value pairs
        """
        for node in tree.get_children():
            yield from self.traverse(node)

            # Assignment
            if isinstance(node, astroid.nodes.Assign):
                if not self.is_value(node.value):
                    continue

                value = self.node_to_str(node.value)
                for key in node.targets:
                    key = self.node_to_str(key)
                    if key and value and isinstance(value, (str, int)):
                        yield KeyValuePair(key, value, line=node.lineno)

            # Comparison
            elif isinstance(node, astroid.nodes.Compare):
                left = node.left
                right = node.ops[0][1]
                key, value = "", ""
                if self.is_key(left) and self.is_value(right):
                    key = self.node_to_str(left)
                    value = self.node_to_str(right)

                elif self.is_key(right) and self.is_value(left):
                    key = self.node_to_str(right)
                    value = self.node_to_str(left)

                if key and value and isinstance(value, (str, int)):
                    yield KeyValuePair(key, value, line=node.lineno)

            # Dictionary values
            elif isinstance(node, astroid.nodes.Dict):
                for key, value in node.items:
                    if not self.is_value(value):
                        continue

                    key = self.node_to_str(key)
                    value = self.node_to_str(value)
                    if key and value and isinstance(value, (str, int)):
                        yield KeyValuePair(key, value, line=node.lineno)

            # Keywords
            elif isinstance(node, astroid.nodes.Keyword):
                key = self.node_to_str(node)
                value = self.node_to_str(node.value)
                if key and value and isinstance(value, (str, int)):
                    yield KeyValuePair(key, value, line=node.lineno)

            # Function call
            elif isinstance(node, astroid.nodes.Call):
                key = "function"
                value = node.as_string()  # Entire function call
                yield KeyValuePair(key, value, line=node.lineno)
                yield from self.parse_env_functions(node)

    def parse_env_functions(self, node: astroid.nodes.Call):
        """
        Decompose environment calls into key-value pairs
        """
        envfuncs = [
            "getenv",
            "environ.get",
        ]
        node_name = node.func.as_string()
        if any(name in node_name for name in envfuncs):
            if len(node.args) == 2:
                key, value = node.args
                key = self.node_to_str(key)
                value = self.node_to_str(value)
                if key and value:
                    yield KeyValuePair(key, value, line=node.lineno)



import logging
from argparse import Namespace
from pathlib import Path

from whispers.core.utils import DEFAULT_PATH, load_yaml_from_file
from whispers.models.appconfig import AppConfig


def load_config(args: Namespace) -> dict:
    """Load config given args"""
    if args.config:
        configfile = Path(args.config)
    else:
        configfile = DEFAULT_PATH.joinpath("config.yml")

    file_exists = False

    try:
        file_exists = configfile.exists()

    except OSError:  # pragma: no cover
        """
        Handle unpredictable file type situations.
        Example:
        Windows NUL behaves differently than Posix /dev/null.
        Doing os.stat(NUL) raises a system error, while
        os.stat(/dev/null) does not.
        """
        raise TypeError(f"{configfile.as_posix()} is not accessible")

    if not file_exists:
        raise FileNotFoundError(f"{configfile.as_posix()} does not exist")

    if not configfile.is_file():
        raise TypeError(f"{configfile.as_posix()} is not a file")

    try:
        config = AppConfig(load_yaml_from_file(configfile))

    except Exception:
        raise RuntimeError(f"{configfile.as_posix()} is invalid")

    logging.debug(f"load_config '{config}'")
    return config



from pathlib import Path
from typing import Iterator

from whispers.core.utils import strip_string
from whispers.models.pair import KeyValuePair
from whispers.plugins.common import Common


class Plaintext:
    def pairs(self, filepath: Path) -> Iterator[KeyValuePair]:
        for lineno, line in enumerate(filepath.open(), 1):
            line = strip_string(line)
            if not line:
                continue

            yield from self.common_pairs(line, lineno)

    def common_pairs(self, text: str, lineno: int) -> Iterator[KeyValuePair]:
        yield from Common(line=lineno).pairs(text)
        yield from self.parse_pk(text)

    @staticmethod
    def parse_pk(text: str) -> Iterator[KeyValuePair]:
        """Check if text resembles a Private Key (PK), only for plaintext files"""
        if not (text.startswith("-----") and text.endswith("-----")):
            return []

        if len(text) < 15:
            return []

        yield KeyValuePair("private_key", text)



from configparser import ConfigParser, MissingSectionHeaderError
from itertools import chain
from pathlib import Path
from typing import Iterator, Optional

from whispers.core.utils import strip_string
from whispers.models.pair import KeyValuePair
from whispers.plugins.xml import Xml


class Config:
    def pairs(self, filepath: Path) -> Iterator[KeyValuePair]:
        if "<?xml " in filepath.open("r").readline():
            yield from Xml().pairs(filepath)

        else:
            try:
                # Attempt to parse as Windows INI
                yield from self.parse_as_ini(filepath)

            except MissingSectionHeaderError:
                # Otherwise, parse as text line by line
                yield from self.parse_as_text(filepath)

    @staticmethod
    def parse_as_ini(filepath: Path) -> Optional[KeyValuePair]:
        parser = ConfigParser()
        parser.read(filepath.as_posix())
        sections = map(lambda section: section.items(), parser.values())
        items = chain.from_iterable(sections)

        for item in items:
            key, value = item
            yield KeyValuePair(key, value)

    @staticmethod
    def parse_as_text(filepath: Path) -> Optional[KeyValuePair]:
        for lineno, line in enumerate(filepath.open(), 1):
            line = line.strip()
            if "=" not in line:
                continue

            key, value = line.split("=", 1)
            key = strip_string(key)
            value = strip_string(value)

            if value:
                yield KeyValuePair(key, value, line=lineno)



import re
from base64 import b64decode
from dataclasses import dataclass, field
from typing import Any, Dict, Optional, Pattern

from whispers.core.utils import is_ascii, is_base64, is_base64_bytes, is_luhn, is_similar, is_uri
from whispers.models.pair import KeyValuePair


@dataclass
class Specification:
    """Rule key/value specification class"""

    regex: Optional[Pattern] = None
    ignorecase: Optional[bool] = False
    minlen: Optional[int] = 0
    isBase64: Optional[bool] = None
    isAscii: Optional[bool] = None
    isUri: Optional[bool] = None
    isLuhn: Optional[bool] = None

    def __post_init__(self) -> None:
        """Compile regex with explicit ignorecase"""
        if self.regex is None:
            return None

        flags = re.IGNORECASE if self.ignorecase else 0
        self.regex = re.compile(self.regex, flags=flags)

    def matches(self, target: str) -> bool:
        if len(target) < self.minlen:
            return False

        if self.regex and not self.regex.match(target):
            return False

        if self.isBase64 is True:
            if is_base64(target) and self.isAscii is not False:
                target = b64decode(target).decode("utf-8")
            elif is_base64_bytes(target) and self.isAscii is False:
                target = b64decode(target)
            else:
                return False

        if self.isAscii is not None and self.isAscii is not is_ascii(target):
            return False

        if self.isUri is not None and self.isUri is not is_uri(target):
            return False

        if self.isLuhn is not None and self.isLuhn is not is_luhn(target):
            return False

        return True


@dataclass
class Rule:
    """Main rule configuration class"""

    id: str = ""
    group: str = ""
    message: str = ""
    severity: str = ""
    key: Specification = field(default=None)
    value: Specification = field(default=None)
    similar: float = field(default_factory=float)

    def __init__(self, rule: Dict) -> None:
        self.id = self._get_required("id", rule)
        self.group = self._get_required("group", rule)
        self.message = self._get_required("message", rule)
        self.severity = self._get_required("severity", rule)
        self.key = self._get_spec("key", rule)
        self.value = self._get_spec("value", rule)
        self.similar = rule.get("similar", 1)

    @staticmethod
    def _get_required(idx: str, rule: Dict) -> Any:
        """Get a required rule value"""
        value = rule.get(idx, False)
        if not value:
            raise IndexError(f"Missing rule '{idx}' specification: '{rule}'")

        return value

    @staticmethod
    def _get_spec(idx: str, rule: Dict) -> Optional[Specification]:
        """Get an optional rule specification"""
        spec = rule.get(idx, False)
        if not spec:
            return None

        if isinstance(spec, dict):
            return Specification(**spec)

        raise ValueError(f"Invalid rule '{idx}' specification: '{rule}'")

    def matches(self, pair: KeyValuePair) -> bool:
        if self.key and not self.key.matches(pair.key):
            return False

        if self.value and not self.value.matches(pair.value):
            return False

        if is_similar(pair.key, pair.value, self.similar):
            return False

        return True



from itertools import chain
from typing import Iterator, List
from urllib.parse import parse_qsl, urlparse

from whispers.models.pair import KeyValuePair


class Common:
    """Checks text for common patterns, such as URI, AWS ARN, etc."""

    def __init__(self, keypath: List = [], line: int = 0) -> None:
        self.keypath = keypath
        self.line = line

    def pairs(self, value: str) -> Iterator[KeyValuePair]:
        """Check for common patterns in text"""
        if not value:
            return []  # Empty

        if not isinstance(value, str):
            return []  # Not a string

        words = value.split(" ")
        uris = map(self.parse_uri, words)
        arns = map(self.parse_arn, words)
        parsed = chain.from_iterable([*uris, *arns])
        yield from parsed

    def parse_uri(self, text: str) -> Iterator[KeyValuePair]:
        """Check if text resembles a Uniform Resource Identifier (URI)"""
        if "://" not in text:
            return []  # Not URI

        uri = urlparse(text)

        if uri.password:
            yield KeyValuePair("uri_creds", f"{uri.username}:{uri.password}", [*self.keypath, text])

        if uri.query:
            for key, value in parse_qsl(uri.query):
                yield KeyValuePair(key, value, [*self.keypath, text], line=self.line)

    def parse_arn(self, text: str) -> Iterator[KeyValuePair]:
        """Check if text resembles an AWS ARN"""
        if not text.startswith("arn:aws:"):
            return []  # Not AWS ARN

        arn = text.split(":")

        if len(arn) < 5 or not arn[4]:
            return []  # Missing AWS Account ID

        account = str(arn[4])

        yield KeyValuePair("aws_account", account, [*self.keypath, text], line=self.line)



from pathlib import Path
from typing import Iterator

from bs4 import BeautifulSoup, Comment

from whispers.core.utils import truncate_all_space
from whispers.models.pair import KeyValuePair


class Html:
    def pairs(self, filepath: Path) -> Iterator[KeyValuePair]:
        soup = BeautifulSoup(filepath.read_text(), "lxml")
        key = "comment"
        for comment in soup.find_all(text=lambda text: isinstance(text, Comment)):
            comment = truncate_all_space(comment.extract()).strip()
            if comment:
                yield KeyValuePair(key, comment, [key])



import logging
import re
import string
from base64 import b64decode
from pathlib import Path
from typing import List, Optional, Pattern, Union

from jellyfish import jaro_winkler_similarity
from luhn import verify as luhn_verify
from yaml import safe_load, safe_load_all

from whispers.models.pair import KeyValuePair

DEFAULT_PATH = Path(__file__).parents[1]
DEFAULT_SEVERITY = ["BLOCKER", "CRITICAL", "MAJOR", "MINOR", "INFO"]

ESCAPED_CHARS = str.maketrans({"'": r"\'", '"': r"\""})

REGEX_URI = re.compile(r"[:\w\d]+://.+", flags=re.IGNORECASE)
REGEX_PATH = re.compile(r"^((([A-Z]|file|root):)?(\.+)?[/\\]+).*$", flags=re.IGNORECASE)
REGEX_IAC = re.compile(r"\![A-Za-z]+ .+", flags=re.IGNORECASE)
REGEX_PRIVKEY_FILE = re.compile(r"(rsa|dsa|ed25519|ecdsa|pem|crt|cer|ca-bundle|p7b|p7c|p7s|ppk|pkcs12|pfx|p12)")
REGEX_ENVVAR = re.compile(r"^\$\$?[A-Z0-9_]+$")


def global_exception_handler(file: Union[str, Path], data: str):
    """Global Exception Handler"""
    logging.exception(f"Failed parsing file '{str(file)}' with {data}")


def load_regex(regex: str, flags: Optional[re.RegexFlag] = 0) -> Pattern:
    """Try to compile a regex statement"""
    try:
        return re.compile(regex, flags=flags)

    except re.error:
        raise ValueError(f"Failed compiling RegEx: {regex}")


def load_yaml_from_file(filepath: Path) -> dict:
    """Safe load yaml from given file path"""
    ret = safe_load(filepath.read_text())
    if not isinstance(ret, dict):
        return {}

    return ret


def truncate_all_space(value: str) -> str:
    """Replace multiple space characters by a single space character"""
    if not value:
        return ""

    return re.sub(r"\s+", " ", value)


def strip_string(value: str) -> str:
    """Strips leading and trailing quotes and spaces"""
    if not value:
        return ""

    return str(value).strip(" '\"\n\r\t")


def simple_string(value: str) -> str:
    """Returns a simplified value for loose comparison"""
    if not value:
        return ""

    value = strip_string(value)  # Remove quotes
    value = value.rstrip("\\")  # Remove trailing backslashes
    value = value.lower()  # Lowercase
    value = re.sub(r"[^a-z0-9]", "_", value.strip())  # Simplify

    return value


def similar_strings(a: str, b: str) -> float:
    """Returns similarity coefficient between two strings"""
    a = simple_string(a).replace("_", "")
    b = simple_string(b).replace("_", "")

    return jaro_winkler_similarity(a, b)


def is_static(key: str, value: str) -> bool:
    """Check if pair is static"""
    if not isinstance(value, str):
        return False  # Not string

    if not value:
        return False  # Empty

    if value.lower() == "null":
        return False  # Empty

    if value.startswith("$") and "$" not in value[2:]:
        if REGEX_ENVVAR.match(value):
            return False  # Variable

    if value.startswith("%") and value.endswith("%"):
        return False  # Variable

    if value.startswith("${") and value.endswith("}"):
        return False  # Variable

    if value.startswith("{") and value.endswith("}"):
        if len(value) > 50:
            if is_base64_bytes(value[1:-1]):
                return True  # Token

        return False  # Variable

    if "{{" in value and "}}" in value:
        return False  # Variable

    if value.startswith("<") and value.endswith(">"):
        return False  # Placeholder

    if value.startswith("ENC[AES256_GCM,data:") and value.endswith("]"):
        return False  # Encrypted SOPS key

    s_key = simple_string(key)
    s_value = simple_string(value)

    if s_key == s_value:
        return False  # Placeholder

    if s_value.endswith(s_key):
        return False  # Placeholder

    if is_iac(value):
        return False  # IaC !Ref !Sub ...

    if is_path(value):
        return False  # System path

    return True  # Hardcoded static value


def is_ascii(data: str) -> bool:
    """Checks if given data is printable text"""
    if isinstance(data, bytes):
        try:
            data = data.decode("utf-8")
        except Exception:
            return False

    if not isinstance(data, (str, int)):
        return False

    for ch in str(data):
        if ch not in string.printable:
            return False

    return True


def is_base64(data: str) -> bool:
    """Checks if given data is base64-decodable to text"""
    if not isinstance(data, str):
        return False

    try:
        b64decode(data).decode("utf-8")
        return True

    except Exception:
        return False


def is_base64_bytes(data: str) -> bool:
    """Checks if given data is base64-decodable to bytes"""
    try:
        return b64decode(data) != b""

    except Exception:
        return False


def is_uri(data: str) -> bool:
    """Checks if given data resemples a URI"""
    if not is_ascii(data):
        return False

    if isinstance(data, int):
        return False

    if any(map(lambda ch: ch in data, string.whitespace)):
        return False

    if not REGEX_URI.match(data):
        return False

    return True


def is_path(data: str) -> bool:
    """Checks if given data resemples a system path"""
    if not is_ascii(data):
        return False

    if isinstance(data, int):
        return False

    if not REGEX_PATH.match(data):
        return False

    return True


def is_iac(data: str) -> bool:
    """Checks if given data resemples IaC function"""
    if not is_ascii(data):
        return False

    if isinstance(data, int):
        return False

    if not REGEX_IAC.match(data):
        return False

    return True


def is_luhn(data: str) -> bool:
    """Checks if given data resembles a credit card number"""
    if not is_ascii(data):
        return False

    if not str(data).isnumeric():
        return False

    return luhn_verify(str(data))


def is_similar(key: str, value: str, similarity: float) -> bool:
    """
    Checks similarity between key and value.
    Returns True if calculated similarity is greater than given similarity.
    """
    return similar_strings(key, value) >= similarity


def find_line_number(pair: KeyValuePair) -> int:
    """Finds line number using pair keypath and value"""
    if pair.line:
        return pair.line  # Already set

    valuepath = pair.value.split("\n")[0][:16]
    findpath = [*pair.keypath, valuepath]
    foundline = 0

    try:
        with Path(pair.file).open() as fh:
            for lineno, line in enumerate(fh, 1):
                founditems = 0

                for item in findpath:
                    if item not in line:
                        break

                    founditems += 1
                    foundline = lineno

                findpath = findpath[founditems:]

                if not findpath:
                    return foundline

    except Exception:  # pragma: no cover
        global_exception_handler(pair.file, "find_line_number()")

    return 0


def default_rules() -> List[dict]:
    """Read and parse builtin rules"""
    rules = []
    files = DEFAULT_PATH.joinpath("rules").glob("*.yml")
    for file in files:
        list(map(rules.extend, safe_load_all(file.read_text())))

    return rules


def list_rule_prop(prop: str, rules: List[dict]) -> List[str]:
    """List rule property given a list of rules"""
    return sorted(set(map(lambda rule: rule[prop], rules)))



from typing import Iterator

from whispers.models.pair import KeyValuePair
from whispers.plugins.common import Common


class StructuredDocument:
    def __init__(self):
        self.keypath = []

    def traverse(self, code, key=None):
        """Recursively traverse YAML/JSON document"""
        if isinstance(code, dict):
            yield from self.cloudformation(code)

            for k, v in code.items():
                self.keypath.append(k)
                if isinstance(v, (str, int)):
                    yield KeyValuePair(k, v, list(self.keypath))

                yield from self.traverse(v, key=k)
                self.keypath.pop()

            # Special key/value format
            elements = list(code.keys())
            if "key" in elements and "value" in elements:
                yield KeyValuePair(code["key"], code["value"], list(self.keypath))

        elif isinstance(code, list):
            for item in code:
                if isinstance(item, (str, int)):
                    yield KeyValuePair(key, item, list(self.keypath))

                yield from self.traverse(item, key=key)

        elif isinstance(code, str):
            if "=" in code:
                item = code.split("=")
                if len(item) == 2:
                    yield KeyValuePair(item[0], item[1], list(self.keypath))

            yield from Common(self.keypath).pairs(code)

    def cloudformation(self, code: dict) -> Iterator[KeyValuePair]:
        """AWS CloudFormation format"""
        if self.keypath:
            return  # Not tree root

        if "AWSTemplateFormatVersion" not in code:
            return  # Not CF format

        if "Parameters" not in code:
            return  # No parameters

        for key, values in code["Parameters"].items():
            if "Default" not in values:
                continue  # No default value

            keypath = ["Parameters", "Default", key]
            yield KeyValuePair(key, values["Default"], keypath)
